from django.conf import settings
from django.core.checks import Error, register, Warning
from django.urls import reverse, NoReverseMatch

from . import app_settings
import environ

@register()
def check_logging_settings(app_configs=None, **kwargs):
    errors = []
    if settings.LOGGING_CONFIG is None:
        errors.append(Warning('LOGGING_CONFIG is set to None which disable logging all together !!'))

    logging_dict = settings.LOGGING
    loggers = logging_dict.get('loggers', {})
    kn_logging = loggers.get('kn_middleware_logger', False)
    if not kn_logging:
        errors.append(Error('`kn_defaults` is not added to LOGGING.loggers',
                            hint="add kn_defaults to LOGGING.loggers objects",
                            obj='settings',
                            id='kn_defaults.E001', ))
    return errors


# @register()
def check_raven(app_configs, **kwargs):
    errors = []
    env = environ.Env()
    sentry_dsn = env.str("SENTRY_DSN", "")

    if not sentry_dsn and not settings.DEBUG:
        errors.append(
            Error('`SENTRY_DSN is not in the environment variables',
                  obj='settings',
                  id='kn_defaults.E002',
                  )
        )
    #
    # raven_config = getattr(settings, 'RAVEN_CONFIG', {})
    # if not raven_config:
    #     errors.append(
    #         Error('`RAVEN_CONFIG` is missing from settings',
    #               hint='Add `RAVEN_CONFIG={"dsn":"<DSN HERE>"}` to relevant settings file',
    #               obj='settings',
    #               id='kn_defaults.E003',
    #               )
    #     )

    return errors


@register()
def check_admin_url(app_configs, **kwargs):
    errors = []
    try:
        admin_url = reverse('admin:index')
    except NoReverseMatch:
        # no default admin in url, exiting
        return errors

    error_class = Warning if settings.DEBUG else Error
    url_parts = admin_url.split('/')
    if 'admin' in url_parts:
        errors.append(
            error_class(
                "it’s not recommended to have the admin url as `/admin/` and it will not work when DEBUG = False. "
                "Please change it to a different url or ask your project manager what url should be used.",
                hint='Change the admin url',
                obj='settings',
                id='kn_defaults.E003')
        )
    return errors


@register()
def check_apm(app_configs, **kwargs):
    errors = []
    # INSTALLATION_NAME = env.list('DJANGO_ALLOWED_HOSTS')[0].replace('.', '-')
    if app_settings.KN_PRODUCTION:
        # check that APM is enabled
        if not getattr(settings, 'ENABLE_APM', False):
            errors.append(Warning('APM is not enabled for this project !!'))
        else:
            if 'elasticapm.contrib.django' not in settings.INSTALLED_APPS:
                errors.append(Error('Please Add APM TO INSTALLED APPS',
                                    hint="INSTALLED_APPS.append('elasticapm.contrib.django')"))
            if 'elasticapm.contrib.django.middleware.TracingMiddleware' not in settings.MIDDLEWARE:
                errors.append(Error('Please Add APM TO MIDDLEWARE',
                                    hint='MIDDLEWARE.insert(0, "elasticapm.contrib.django.middleware.TracingMiddleware")'))
            ELASTIC_APM = getattr(settings, 'ELASTIC_APM', False)
            if not ELASTIC_APM:
                errors.append(Error('Please Add ELASTIC_APM TO your settings file',
                                    hint="""ELASTIC_APM = {
    "SERVICE_NAME": INSTALLATION_NAME,
    "SERVER_URL": "<YOUR_UP:PORT>",
    "DJANGO_TRANSACTION_NAME_FROM_ROUTE": True,
}"""))


            if not getattr(settings, 'INSTALLATION_NAME', False):
                errors.append(Error('Please Add an APM installation name',
                                    hint="INSTALLATION_NAME = env.list('DJANGO_ALLOWED_HOSTS')[0].replace('.', '-')"))
    return errors



from django.conf import settings
from django.utils.functional import lazy
import os

KN_LOGGING_URL_PATTERNS = lazy(lambda: getattr(settings, 'KN_LOGGING_URL_PATTERNS', []), list)()
DISABLE_CMS_PLUGIN_CHANGE_ADMIN_LOG = lazy(lambda: getattr(settings, 'DISABLE_CMS_PLUGIN_CHANGE_ADMIN_LOG', False),
                                           bool)()
KN_LOG_FILE_SIZE = lazy(lambda: getattr(settings, 'KN_LOG_FILE_SIZE', 5 * 1024 * 1024), int)()
KN_HANDLER_CLASS = lazy(
    lambda: getattr(settings, 'KN_HANDLER_CLASS', 'logging.handlers.RotatingFileHandler'), str)()
KN_LOG_FILE_PATH = lazy(
    lambda: getattr(settings, 'KN_LOG_FILE_PATH', os.path.join(os.getcwd(), 'log.log')), str)()
KN_LOG_BACKUP_COUNT = lazy(lambda: getattr(settings, 'KN_LOG_BACKUP_COUNT', 3), int)()
KN_PRODUCTION = lazy(lambda: getattr(settings, 'KN_PRODUCTION', not settings.DEBUG), bool)()



import sys
from functools import wraps
import logging

logger = logging.getLogger('kn_defaults')
from django.utils.decorators import method_decorator


def log_activity(verb, target=None, level=0):
    def decorator(func):
        @wraps(func)
        def inner(self, *args, **kwargs):
            obj = func(self, *args, **kwargs)
            # if hasattr(self, 'txn_type'):
            #     # means the object is PaymentTransaction
            #     # get the verbose name of it.
            #     msg = f'{verb} ' \
            #         f'{next(f[1] for f in constants.TYPE_CHOICES if f[0] == self.txn_type)}'
            # else:
            msg = verb

            params = {'verb': msg}

            if callable(target):
                params['target'] = target(obj)

            # sometimes target object is passed in, because action_object doesn't exist,
            # ie: it was deleted
            # in that case, obj returned by the function is the target obj
            if obj is not None and params.get('target') != obj:
                params['action_object'] = obj

            logger.log(level, f'{params["verb"]} - f{params["target"]}', exc_info=sys.exec_info, extra={
                'request': self.request,
                'status_code': obj.status_code
            })
            # action.send(self.request.user, **params)

            return obj

        return inner

    return decorator





from django import apps

from django.conf import settings


class KnDefault(apps.AppConfig):
    name = 'kn_defaults.logging'
    verbose_name = 'KuwaitNet Helpers'

    def ready(self):
        super().ready()
        from .app_settings import DISABLE_CMS_PLUGIN_CHANGE_ADMIN_LOG
        from . import checks
        if 'cms' in settings.INSTALLED_APPS and not DISABLE_CMS_PLUGIN_CHANGE_ADMIN_LOG:
            from .handlers import cms_plugin_change_admin_log
            from cms.signals import post_placeholder_operation
            post_placeholder_operation.connect(cms_plugin_change_admin_log)



from functools import wraps
import logging
import os
import environ

env = environ.Env()

logger = logging.getLogger('default')

KN_FORMATTER = "%(levelname)s:%(name)s; " \
               "REQ_id:%(request_id)s; %(message)s; path=%(path)s; method=%(method)s;ip=%(ip)s; " \
               "status_code:%(status_code)s; response_duration:%(response_duration)s; " \
               "post_parameters: %(post_parameters)s; outbound:%(outbound_payload)s; META: %(meta)s"

FUNCTION_LOGGER_FORMATTER = '{levelname}:{message} - args={func_args} kwargs={func_kwargs} return={func_return_value} '
PROJECT_NAME = env.str('DJANGO_PROJECT_NAME')
PROJECT_ROOT = env.str('DJANGO_PROJECT_ROOT')
LOGSTASH_HOST = env.str('DJANGO_LOGSTASH_HOST')
LOGSTASH_PORT = env.int('DJANGO_LOGSTASH_PORT')

LOGSTASH_ENV = env.str('DJANGO_LOGSTASH_ENV', 'dev')
LOGSTASH_EXTRA_PREFIX = env.str('DJANGO_LOGSTASH_EXTRA_PREFIX', 'dev')
LOGSTASH_SSL_ENABLE = env.bool('DJANGO_LOGSTASH_SSL_ENABLE', False)

BASE_LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        # 'logstash-verbose_middleware': {
        #     # 'format': KN_FORMATTER,
        #     "()": "logstash_async.formatter.DjangoLogstashFormatter",
        #     "message_type": "python-logstash",
        #     "fqdn": False,  # Fully qualified domain name. Default value: false.
        #     'extra_prefix': LOGSTASH_EXTRA_PREFIX,
        #     "extra": {
        #         "application": PROJECT_NAME,
        #         "project_path": PROJECT_ROOT,
        #         "environment": LOGSTASH_ENV,
        #     },
        # },
        # 'logstash-verbose_project': {
        #     # 'format': '{levelname}:{name}:{message} - LocalVars={vars} ',
        #     # 'style': '{',
        #     "()": "logstash_async.formatter.DjangoLogstashFormatter",
        #     "message_type": "python-logstash",
        #     "fqdn": False,
        #     'extra_prefix': LOGSTASH_EXTRA_PREFIX,
        #     "extra": {
        #         "application": PROJECT_NAME,
        #         "project_path": PROJECT_ROOT,
        #         "environment": LOGSTASH_ENV,
        #     },
        # },
        # 'logstash-verbose_functions': {
        #     # 'format': FUNCTION_LOGGER_FORMATTER,
        #     # 'style': '{',
        #     "()": "logstash_async.formatter.DjangoLogstashFormatter",
        #     "message_type": "python-logstash",
        #     "fqdn": False,
        #     'extra_prefix': LOGSTASH_EXTRA_PREFIX,
        #     "extra": {
        #         "application": PROJECT_NAME,
        #         "project_path": PROJECT_ROOT,
        #         "environment": LOGSTASH_ENV,
        #     },
        # },
        "logstash": {
            "()": "logstash_async.formatter.DjangoLogstashFormatter",
            "message_type": "python-logstash",
            "fqdn": False,
            'extra_prefix': LOGSTASH_EXTRA_PREFIX,
            "extra": {
                "application": PROJECT_NAME,
                "project_path": PROJECT_ROOT,
                "environment": LOGSTASH_ENV,
            },
        }
    },
    'handlers': {
        # 'kn_default_handler': {
        #     'class': 'logging.StreamHandler',
        #     'formatter': 'verbose_middleware'
        # },
        # 'middleware_handler': {
        #     'level': 'DEBUG',
        #     "class": "logstash_async.handler.AsynchronousLogstashHandler",
        #     "transport": "logstash_async.transport.TcpTransport",
        #     "host": LOGSTASH_HOST,
        #     "port": LOGSTASH_PORT,
        #     "ssl_enable": LOGSTASH_SSL_ENABLE,
        #     "database_path": f"{PROJECT_ROOT}/logstash.db",
        #     'formatter': 'logstash-verbose_middleware',
        # },
        # 'file_log': {
        #     'level': 'DEBUG',
        #     "class": "logstash_async.handler.AsynchronousLogstashHandler",
        #     "transport": "logstash_async.transport.TcpTransport",
        #     "host": LOGSTASH_HOST,
        #     "port": LOGSTASH_PORT,
        #     "ssl_enable": LOGSTASH_SSL_ENABLE,
        #     "database_path": f"{PROJECT_ROOT}/logstash.db",
        #     'formatter': 'logstash-verbose_project',
        # },
        # 'functions_log': {
        #     'level': 'DEBUG',
        #     "class": "logstash_async.handler.AsynchronousLogstashHandler",
        #     "transport": "logstash_async.transport.TcpTransport",
        #     "host": LOGSTASH_HOST,
        #     "port": LOGSTASH_PORT,
        #     "ssl_enable": LOGSTASH_SSL_ENABLE,
        #     "database_path": f"{PROJECT_ROOT}/logstash.db",
        #     'formatter': 'logstash-verbose_functions',
        # },
        "logstash": {
            "level": "DEBUG",
            "class": "logstash_async.handler.AsynchronousLogstashHandler",
            "formatter": "logstash",
            "transport": "logstash_async.transport.TcpTransport",
            "host": LOGSTASH_HOST,
            "port": LOGSTASH_PORT,
            "ssl_enable": LOGSTASH_SSL_ENABLE,
            "database_path": f"{PROJECT_ROOT}/logstash.db",
        },
    },
    'loggers': {
        'kn_middleware_logger': {
            'handlers': ['logstash'],
            'level': 'INFO',
        },
        'default': {
            'handlers': ['logstash'],
            'level': 'DEBUG',
        },
        'kn_function_logger': {
            'handlers': ['logstash'],
            'level': 'DEBUG',
        }
    }
}


def log(level, msg, collect_localvars=True, exc_info=None, extra=None, stack_info=False):
    import inspect
    extra = extra or {}
    vars = {}
    caller_name = ''
    if collect_localvars:
        frame = inspect.currentframe()
        calframe = inspect.getouterframes(frame, 2)
        try:
            vars = frame.f_back.f_locals
            caller_name = calframe[1][3]

        finally:
            del frame

    extra['vars'] = vars
    extra['caller_name'] = vars

    logger.log(level, msg, exc_info=exc_info, extra=extra, stack_info=stack_info)


def logging_decorator(func, *, level=10, msg=''):
    @wraps(func)
    def function_wrapper(*args, **kwargs):
        logger = logging.getLogger('kn_function_logger')
        message = msg or func.__name__

        return_value = func(*args, **kwargs)
        logger.log(level, message, extra={'func_args': args, 'func_kwargs': kwargs, 'func_return_value': return_value})
        return return_value

    return function_wrapper
